/* MoreBalance - load balancer and networking tool
 * Copyright (C) 2002 - 2006 Josef Spillner <josef@coolprojects.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* MoreBalance includes */
#include "conf.h"
#include "sys.h"
#include "spec.h"

/* System includes */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>

/* Parser states */
#define STATE_START 0

#define STATE_PROTO        1
#define STATE_POSTPROTO    2
#define STATE_NODE         3
#define STATE_POSTNODE     4
#define STATE_AS           5
#define STATE_MATCHING     6
#define STATE_MATCHING2    7
#define STATE_LAUNCHING    8
#define STATE_WITH         9

#define STATE_PROTO1      11
#define STATE_POSTPROTO1  12
#define STATE_PROTO2      13
#define STATE_POSTPROTO2  14
#define STATE_TIME        15
#define STATE_TIMEUNIT    16

#define STATE_BE          21
#define STATE_VERY        22

#define STATE_ACCEPT      31
#define STATE_POSTACCEPT  32
#define STATE_CONNECTIONS 33
#define STATE_IN          34
#define STATE_EACH        35
#define STATE_POSTEACH    36
#define STATE_FROM        37
#define STATE_POSTFROM    38

#define STATE_STOP       100

/* Globals */
static conf c;
static int initialized = 0;
static int linenumber = 0;

/* Give an detailed error message */
static conf error(const char *buffer, int line, int position, const char *reason)
{
	int pos, num;

	c.valid = 0;
	c.verbose = 0;
	pos = 0;
	num = 0;

	fprintf(stderr, _("ERROR: configuration error in line %i\n"), line);
	fprintf(stderr, "%s\n", buffer);
	while((num < position) && (pos < strlen(buffer)))
	{
		if(buffer[pos] == ' ') num++;
		pos++;
		fprintf(stderr, "-");
	}
	fprintf(stderr, _("^ here.\n"));
	fprintf(stderr, _("Possible reason: %s.\n"), reason);

	return c;
}

/* Reset the configuration object */
static void conf_reset(int once)
{
	c.valid = 1;
	c.verbose = 0;
	c.cluster = NULL;
	c.qos = NULL;
	c.qosnum = 0;
	c.clusternum = 0;

	initialized = once;
	linenumber = 0;
}

/* Read the configuration from a file */
conf conf_read(const char *filename)
{
	FILE *f;
	char buf[1024];

	f = fopen(filename, "r");
	if(!f) return c;

	conf_reset(1);

	/* Read every single line */
	while(fgets(buf, sizeof(buf), f))
	{
		linenumber++;

		buf[strlen(buf) - 1] = 0;
		if((!strlen(buf) || (buf[0] == '#'))) continue;

		c = conf_readstring(buf);
		if(!c.valid) break;
	}
	fclose(f);

	initialized = 0;

	return c;
}

conf conf_readstring(const char *string)
{
	char *pretoken, *token, *tokenbuffer, *pretokenbuffer;
	char *str, *orig;
	int state;
	int port;
	int counter, xcounter;
	long time, conn;
	cluster *ctmp;
	qos *qtmp;
	int reverse;
	int istoken;

	orig = strdup(string);
	str = strdup(string);
	pretoken = strtok_r(str, "\"", &pretokenbuffer);
	counter = 0;
	xcounter = 0;
	state = STATE_START;
	reverse = 0;

	time = 0;
	conn = 0;
	qtmp = NULL;
	ctmp = NULL;

	if(!initialized)
	{
		conf_reset(0);
	}

	/* Tokenize each line */
	while(pretoken)
	{
		if(orig[xcounter] != '\"')
		{
			token = strtok_r(pretoken, " ", &tokenbuffer);
			istoken = 0;
		}
		else
		{
			token = pretoken;
			istoken = 1;
		}
		while(token)
		{
			xcounter += strlen(token) + 1;
			if((state == STATE_START) || (state == STATE_PROTO) || (state == STATE_PROTO1) || state == (STATE_PROTO2))
			{
				if((state == STATE_START) && (!strcmp(token, "guarantee")))
				{
					qtmp = (qos*)malloc(sizeof(qos));
					state = STATE_PROTO1;
				}
				else if((state == STATE_START) && (!strcmp(token, "be")))
				{
					state = STATE_BE;
				}
				else if((state == STATE_START) && (!strcmp(token, "accept")))
				{
					state = STATE_ACCEPT;
				}
				else
				{
					port = sys_port(token);
					if(port == 0)
					{
						if(token[0] != '/')
							return error(orig, linenumber, counter, "wrong service");
					}
					if((state == STATE_PROTO) || (state == STATE_START))
					{
						if(state == STATE_START)
						{
							ctmp = (cluster*)malloc(sizeof(cluster));
							ctmp->ports = NULL;
							ctmp->portopts = NULL;
							ctmp->nodes = NULL;
							ctmp->sockets = NULL;
							ctmp->nodesnum = 0;
							ctmp->portsnum = 0;
							ctmp->destport = 0;
							ctmp->destportopts = 0;
							ctmp->destsocket = NULL;
							ctmp->pattern = NULL;
							ctmp->prelaunch = NULL;
						}

						ctmp->ports = (int*)realloc(ctmp->ports,
							(ctmp->portsnum + 1) * sizeof(int));
						ctmp->ports[ctmp->portsnum] = port;
						ctmp->portopts = (int*)realloc(ctmp->portopts,
							(ctmp->portsnum + 1) * sizeof(int));
						ctmp->portopts[ctmp->portsnum] = sys_portoptions(token);
						ctmp->sockets = (char**)realloc(ctmp->sockets,
							(ctmp->portsnum + 1) * sizeof(char*));
						ctmp->sockets[ctmp->portsnum] = (port == 0 ? strdup(token) : NULL);
						ctmp->portsnum++;

						state = STATE_POSTPROTO;
					}
					else if(state == STATE_PROTO1)
					{
						qtmp->port1 = port;
						state = STATE_POSTPROTO1;
					}
					else if(state == STATE_PROTO2)
					{
						qtmp->port2 = port;

						if(reverse)
						{
							port = qtmp->port1;
							qtmp->port1 = qtmp->port2;
							qtmp->port2 = port;
						}

						state = STATE_POSTPROTO2;
					}
				}
			}
			else if(state == STATE_ACCEPT)
			{
				conn = strtol(token, NULL, 10);
				if((conn == LONG_MIN) || (conn == LONG_MAX))
					return error(orig, linenumber, counter, _("number needed"));
				state = STATE_POSTACCEPT;
			}
			else if(state == STATE_POSTACCEPT)
			{
				if(!strcmp(token, "connections")) state = STATE_CONNECTIONS;
				else return error(orig, linenumber, counter, _("unknown token; 'connections' expected"));
			}
			else if(state == STATE_CONNECTIONS)
			{
				if(!strcmp(token, "in")) state = STATE_IN;
				else if(!strcmp(token, "each"))
				{
					time = 1;
					state = STATE_EACH;
				}
				else return error(orig, linenumber, counter, _("unknown token; 'in' or 'each' expected"));
			}
			else if(state == STATE_IN)
			{
				time = strtol(token, NULL, 10);
				if((time == LONG_MIN) || (time == LONG_MAX))
					return error(orig, linenumber, counter, _("number needed"));
				state = STATE_EACH;
			}
			else if(state == STATE_EACH)
			{
				if((!strcmp(token, "second")) || (!strcmp(token, "seconds")))
					time = time;
				else if((!strcmp(token, "minute")) || (!strcmp(token, "minutes")))
					time = time * 60;
				else if((!strcmp(token, "hour")) || (!strcmp(token, "hours")))
					time = time * 3600;
				else return error(orig, linenumber, counter, "unkown time unit");

				state = STATE_POSTEACH;
			}
			else if(state == STATE_POSTEACH)
			{
				if(!strcmp(token, "from")) state = STATE_FROM;
				else return error(orig, linenumber, counter, _("unknown token; 'from' expected"));
			}
			else if(state == STATE_FROM)
			{
				/* ... */
				state = STATE_STOP;
			}
			else if((state == STATE_BE) || (state == STATE_VERY))
			{
				if(!strcmp(token, "verbose"))
				{
					c.verbose += 1;
					state = STATE_STOP;
				}
				else if((state == STATE_BE) && (!strcmp(token, "very")))
				{
					c.verbose += 1;
					state = STATE_VERY;
				}
				else return error(orig, linenumber, counter, _("unknown token; 'verbose' or 'very' expected"));
			}
			else if(state == STATE_POSTPROTO)
			{
				if(!strcmp(token, "and")) state = STATE_PROTO;
				else if(!strcmp(token, "use")) state = STATE_NODE;
				else return error(orig, linenumber, counter, _("unknown token; 'and' or 'use' expected"));
			}
			else if(state == STATE_NODE)
			{
				ctmp->nodes = (char**)realloc(ctmp->nodes, (ctmp->nodesnum + 1 * sizeof(char*)));
				ctmp->nodes[ctmp->nodesnum] = strdup(token);
				ctmp->nodesnum++;

				state = STATE_POSTNODE;
			}
			else if(state == STATE_AS)
			{
				port = sys_port(token);
				if(port == 0)
				{
					if(token[0] != '/')
						return error(orig, linenumber, counter, _("wrong service"));
				}
				ctmp->destport = port;
				ctmp->destportopts = sys_portoptions(token);
				ctmp->destsocket = (port == 0 ? strdup(token) : NULL);
				//state = STATE_PROTO3;
				state = STATE_POSTNODE;
			}
			else if(state == STATE_WITH)
			{
				if(!strcmp(token, "compression"))
				{
					/* discard - should apply to all nodes */
				}
				else if(!strcmp(token, "decompression"))
				{
					/* discard - should apply to all nodes */
				}
				else return error(orig, linenumber, counter,
					_("unknown token; 'compression' or 'decompression' expected"));
			}
			else if(state == STATE_POSTNODE)
			{
				if(!strcmp(token, "and")) state = STATE_NODE;
				else if(!strcmp(token, "as")) state = STATE_AS;
				else if(!strcmp(token, "when")) state = STATE_MATCHING;
				else if(!strcmp(token, "launching")) state = STATE_LAUNCHING;
				else if(!strcmp(token, "with")) state = STATE_WITH;
				else return error(orig, linenumber, counter,
					_("unknown token; 'and' or 'as' or 'when' or 'launching' or 'with' expected"));
			}
			else if(state == STATE_POSTPROTO1)
			{
				if(!strcmp(token, "before"))
				{
					reverse = 1;

					state = STATE_PROTO2;
				}
				else if(!strcmp(token, "after")) state = STATE_PROTO2;
				else return error(orig, linenumber, counter,
					_("unknown token; 'before' or 'after' expected"));
			}
			else if(state == STATE_POSTPROTO2)
			{
				if(!strcmp(token, "for")) state = STATE_TIME;
				else return error(orig, linenumber, counter, _("unknown token; 'for' expected"));
			}
			else if(state == STATE_LAUNCHING)
			{
				ctmp->prelaunch = strdup(token);
				state = STATE_POSTNODE;
			}
			else if(state == STATE_MATCHING)
			{
				if(!strcmp(token, "matching")) state = STATE_MATCHING2;
				else return error(orig, linenumber, counter, _("unknown token; 'matching' expected"));
			}
			else if(state == STATE_MATCHING2)
			{
				/*printf("*** match %s ***\n", token);*/
				ctmp->pattern = strdup(token);
				state = STATE_POSTNODE;
			}
			else if(state == STATE_TIME)
			{
				time = strtol(token, NULL, 10);
				if((time == LONG_MIN) || (time == LONG_MAX))
					return error(orig, linenumber, counter, _("number needed"));
				state = STATE_TIMEUNIT;
			}
			else if(state == STATE_TIMEUNIT)
			{
				if(!strcmp(token, "seconds"))
					time = time;
				else if(!strcmp(token, "minutes"))
					time = time * 60;
				else if(!strcmp(token, "hours"))
					time = time * 3600;
				else return error(orig, linenumber, counter, _("unkown time unit"));
				qtmp->timeout = time;

				c.qos = (qos**)realloc(c.qos, (c.qosnum + 1) * sizeof(qos*));
				c.qos[c.qosnum] = qtmp;
				c.qosnum++;

				state = STATE_STOP;
			}
			else if(state == STATE_STOP)
			{
				return error(orig, linenumber, counter, _("trailing invalid text"));
			}
			else return error(orig, linenumber, counter, _("unknown token"));

			if(!istoken) token = strtok_r(NULL, " ", &tokenbuffer);
			else token = NULL;
			counter++;
		}

		pretoken = strtok_r(NULL, "\"", &pretokenbuffer);
		counter++;
	}

	/* Handle unfinished states */
	if(state != STATE_STOP)
	{
		if(state == STATE_POSTPROTO2)
		{
			qtmp->timeout = 60;

			c.qos = (qos**)realloc(c.qos, (c.qosnum + 1) * sizeof(qos*));
			c.qos[c.qosnum] = qtmp;
			c.qosnum++;
		}
		else if(state == STATE_POSTNODE)
		{
			c.cluster = (cluster**)realloc(c.cluster, (c.clusternum + 1) * sizeof(cluster*));
			c.cluster[c.clusternum] = ctmp;
			c.clusternum++;
		}
		else if(state == STATE_POSTEACH)
		{
			/* ... */
		}
		else if(state == STATE_NODE) return error(orig, linenumber, counter, _("node name expected"));
		else if((state == STATE_PROTO) || (state == STATE_PROTO1) || (state == STATE_PROTO2))
		{
			return error(orig, linenumber, counter, _("need to specify a service here"));
		}
		else return error(orig, linenumber, counter, _("missing end of configuration directive"));
	}

	free(orig);
	free(str);

	return c;
}

/* Check for validity */
void conf_check(conf c)
{
	int i, j, k;
	int ret;

	for(i = 0; i < c.clusternum; i++)
	{
		for(j = 0; j < c.cluster[i]->portsnum; j++)
		{
#ifndef USE_ZLIB
			if((c.cluster[i]->portopts[j] & opt_compress) || (c.cluster[i]->destportopts & opt_compress))
				printf(_("Warning: Compression is not supported.\n"));
#endif
			if((c.cluster[i]->portopts[j] & opt_compress) && (c.cluster[i]->destportopts & opt_compress))
				printf(_("Warning: Compression on both ends makes no sense.\n"));
#ifndef USE_TLS
			if(c.cluster[i]->portopts[j] & opt_tls)
				printf(_("Warning: TLS is not supported.\n"));
			if(c.cluster[i]->destportopts & opt_tls)
				printf(_("Warning: TLS is not supported.\n"));
#endif
			if((c.cluster[i]->portopts[j] & opt_tls) && (c.cluster[i]->portopts[j] & opt_udp))
				printf(_("Warning: TLS cannot be used on UDP connections.\n"));
			if((c.cluster[i]->destportopts & opt_tls) && (c.cluster[i]->destportopts & opt_udp))
				printf(_("Warning: TLS cannot be used on UDP connections.\n"));

			if((c.cluster[i]->portopts[j] & opt_udp) && (!(c.cluster[i]->destportopts & opt_udp)))
				printf(_("Warning: Connection between UDP and TCP endpoints.\n"));
			if((!(c.cluster[i]->portopts[j] & opt_udp)) && (c.cluster[i]->destportopts & opt_udp))
				printf(_("Warning: Connection between UDP and TCP endpoints.\n"));

			for(k = 0; k < c.cluster[i]->nodesnum; k++)
			{
				ret = sys_local(c.cluster[i]->nodes[k]);
				if(ret < 0)
					printf(_("Warning: Host %s can currently not be found\n"),
						c.cluster[i]->nodes[k]);
			}

			if(c.cluster[i]->prelaunch)
			{
				for(k = 0; k < c.cluster[i]->nodesnum; k++)
				{
					ret = sys_local(c.cluster[i]->nodes[k]);
					if(!ret)
						printf(_("Warning: Launching program %s on remote host.\n"),
							c.cluster[i]->prelaunch);
				}
			}

			if((c.cluster[i]->destport) || (c.cluster[i]->destsocket))
			{
				if(!c.cluster[i]->destport)
					for(k = 0; k < c.cluster[i]->nodesnum; k++)
					{
						ret = sys_local(c.cluster[i]->nodes[k]);
						if(!ret)
							printf(_("Warning: Socket %s is on remote host %s.\n"),
								c.cluster[i]->destsocket, c.cluster[i]->nodes[k]);

					}
				if(c.cluster[i]->destportopts & opt_local)
					printf(_("Warning: Option 'local' makes no sense for destination.\n"));
			}
		}
	}
}

/* Dump the whole configuration */
void conf_dump(conf c)
{
	int i, j, k;

	for(i = 0; i < c.qosnum; i++)
	{
		printf(_("* QoS: %i after %i [%li seconds]\n"), c.qos[i]->port1, c.qos[i]->port2, c.qos[i]->timeout);
	}
	for(i = 0; i < c.clusternum; i++)
	{
		for(j = 0; j < c.cluster[i]->portsnum; j++)
		{
			printf(_("* Cluster: "));
			for(k = 0; k < c.cluster[i]->nodesnum; k++)
			{
				printf("%s ", c.cluster[i]->nodes[k]);
			}
			if(c.cluster[i]->ports[j])
				printf(_(" [port %i"), c.cluster[i]->ports[j]);
			else
				printf(_(" [socket %s"), c.cluster[i]->sockets[j]);
			if(c.cluster[i]->portopts[j] & opt_tls) printf(" (tls)");
			if(c.cluster[i]->portopts[j] & opt_udp) printf(" (udp)");
			if(c.cluster[i]->portopts[j] & opt_tcp) printf(" (tcp)");
			if(c.cluster[i]->portopts[j] & opt_compress) printf(" (compress)");
			if(c.cluster[i]->portopts[j] & opt_local) printf(" (local)");
			printf("]");
			if((c.cluster[i]->destport) || (c.cluster[i]->destsocket))
			{
				if(c.cluster[i]->destport)
					printf(_(" [-> port %i"), c.cluster[i]->destport);
				else
					printf(_(" [-> socket %s"), c.cluster[i]->destsocket);
				if(c.cluster[i]->destportopts & opt_tls) printf(" (tls)");
				if(c.cluster[i]->destportopts & opt_udp) printf(" (udp)");
				if(c.cluster[i]->destportopts & opt_tcp) printf(" (tcp)");
				if(c.cluster[i]->destportopts & opt_compress) printf(" (compress)");
				if(c.cluster[i]->destportopts & opt_local) printf(" (local)");
				printf("]");
			}
			printf("\n");
		}
	}
}

/* Optimize the configuration set */
void conf_optimize(conf *c)
{
	int i, j, k, l, x;
	cluster *ctmp, *tmp;

	/* Erase duplicate entries in each cluster */
	for(i = 0; i < c->clusternum; i++)
	{
		tmp = c->cluster[i];

		for(k = 0; k < tmp->portsnum; k++)
		{
			for(l = k + 1; l < tmp->portsnum; l++)
			{
				if(tmp->ports[k] == tmp->ports[l])
				{
					if((tmp->ports[k]) || ((tmp->sockets[k]) && (!strcmp(tmp->sockets[k], tmp->sockets[l]))))
					{
						/* Remove port from one entry */
						if(tmp->sockets[l])
							free(tmp->sockets[l]);
						for(x = l + 1; x < tmp->portsnum; x++)
						{
							tmp->ports[x - 1] = tmp->ports[x];
							tmp->sockets[x - 1] = tmp->sockets[x];
						}
						tmp->portsnum--;
						tmp->ports = (int*)realloc(tmp->ports, (tmp->portsnum * sizeof(int)));
						tmp->sockets = (char**)realloc(tmp->sockets, (tmp->portsnum * sizeof(char*)));

						l--;
					}
				}
			}
		}
	}

	/* Merge cluster entries with same ports into a new dedicated entry */
	for(i = 0; i < c->clusternum; i++)
	{
		for(j = i + 1; j < c->clusternum; j++)
		{
			for(k = 0; k < c->cluster[i]->portsnum; k++)
			{
				for(l = 0; l < c->cluster[j]->portsnum; l++)
				{
					if(c->cluster[i]->ports[k] == c->cluster[j]->ports[l])
					{
						/* Create new cluster */
						ctmp = (cluster*)malloc(sizeof(cluster));
						ctmp->portsnum = 1;
						ctmp->ports = (int*)malloc(sizeof(int));
						ctmp->ports[0] = c->cluster[i]->ports[k];
						if(c->cluster[i]->sockets[k])
							ctmp->sockets[0] = strdup(c->cluster[i]->sockets[k]);
						else
							ctmp->sockets[0] = NULL;
						ctmp->nodesnum = c->cluster[i]->nodesnum + c->cluster[j]->nodesnum;
						ctmp->nodes = (char**)malloc(ctmp->nodesnum * sizeof(char*));
						for(x = 0; x < c->cluster[i]->nodesnum; x++)
						{
							ctmp->nodes[x] = strdup(c->cluster[i]->nodes[x]);
						}
						for(x = 0; x < c->cluster[j]->nodesnum; x++)
						{
							ctmp->nodes[x + c->cluster[i]->nodesnum] = strdup(c->cluster[j]->nodes[x]);
						}
						if(ctmp->pattern)
							ctmp->pattern = strdup(c->cluster[i]->pattern);
						else
							ctmp->pattern = NULL;
						ctmp->destport = c->cluster[i]->destport;
						ctmp->destportopts = c->cluster[i]->destportopts;
						if(c->cluster[i]->destsocket)
							ctmp->destsocket = strdup(c->cluster[i]->destsocket);
						else
							ctmp->destsocket = NULL;
						ctmp->portopts = (int*)malloc(sizeof(int));
						ctmp->portopts[0] = c->cluster[i]->portopts[k];
						if(c->cluster[i]->prelaunch)
							ctmp->prelaunch = strdup(c->cluster[i]->prelaunch);
						else
							ctmp->prelaunch = NULL;

						/* Insert into cluster list */
						c->clusternum++;
						c->cluster = (cluster**)realloc(c->cluster, c->clusternum * sizeof(cluster*));
						c->cluster[c->clusternum - 1] = ctmp;

						/* Remove ports from original two entries */
						if(c->cluster[i]->sockets[k])
							free(c->cluster[i]->sockets[k]);
						for(x = k + 1; x < c->cluster[i]->portsnum; x++)
						{
							c->cluster[i]->ports[x - 1] = c->cluster[i]->ports[x];
							c->cluster[i]->sockets[x - 1] = c->cluster[i]->sockets[x];
						}
						c->cluster[i]->portsnum--;
						c->cluster[i]->ports = (int*)realloc(c->cluster[i]->ports, (c->cluster[i]->portsnum * sizeof(int)));
						c->cluster[i]->sockets = (char**)realloc(c->cluster[i]->sockets, (c->cluster[i]->portsnum * sizeof(char*)));

						if(c->cluster[j]->sockets[l])
							free(c->cluster[j]->sockets[l]);
						for(x = l + 1; x < c->cluster[j]->portsnum; x++)
						{
							c->cluster[j]->ports[x - 1] = c->cluster[j]->ports[x];
							c->cluster[j]->sockets[x - 1] = c->cluster[j]->sockets[x];
						}
						c->cluster[j]->portsnum--;
						c->cluster[j]->ports = (int*)realloc(c->cluster[j]->ports, (c->cluster[j]->portsnum * sizeof(int)));
						c->cluster[j]->sockets = (char**)realloc(c->cluster[j]->sockets, (c->cluster[j]->portsnum * sizeof(char*)));

						k--;
						l--;
					}
				}
			}
		}
	}

	/* Delete empty clusters (no more ports registered) */
	for(i = 0; i < c->clusternum; i++)
	{
		tmp = c->cluster[i];

		if(tmp->portsnum == 0)
		{
			for(j = 0; j < tmp->nodesnum; j++)
			{
				free(tmp->nodes[j]);
			}
			free(tmp->nodes);
			free(tmp->ports);
			free(tmp->portopts);
			free(tmp->sockets);
			if(tmp->pattern) free(tmp->pattern);
			if(tmp->prelaunch) free(tmp->prelaunch);
			if(tmp->destsocket) free(tmp->destsocket);
			free(tmp);

			for(j = i + 1; j < c->clusternum; j++)
			{
				c->cluster[j - 1] = c->cluster[j];
			}
			c->clusternum--;
			c->cluster = (cluster**)realloc(c->cluster, c->clusternum * sizeof(cluster*));

			i--;
		}
	}
}

