/* MoreBalance - load balancer and networking tool
 * Copyright (C) 2002 - 2006 Josef Spillner <josef@coolprojects.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "ipban.h"
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <pthread.h>

/* Seconds of delay between connection attempts */
#define IPBAN_TARPIT 2
/* Seconds after which the ban level is dropped */
#define IPBAN_RESET 1800
/* Minimum level for being banned */
#define IPBAN_BANNED 1000

/* Internal types */
struct ipban_t
{
	unsigned int base;
	unsigned int range;
	time_t nexttry;
	unsigned int banlevel;
};
typedef struct ipban_t ipban;

/* Global variables */
static ipban **banlist = NULL;
static int bansize = 0;
static ipban tmp;

/* Helper function: power(base, exp) */
static unsigned int _pow(unsigned int base, unsigned int power)
{
	unsigned int ret = 1;
	int i;

	for(i = 0; i < power; i++)
		ret *= base;

	return ret;
}

/* Internal function: initialize list with default entry */
static void _ipban_init()
{
	banlist = (ipban**)malloc(sizeof(ipban*));
	banlist[0] = (ipban*)malloc(sizeof(ipban));
	banlist[0]->base = 0;
	banlist[0]->range = 0;
	banlist[0]->banlevel = 1;
	banlist[0]->nexttry = time(NULL) + IPBAN_TARPIT;
	bansize = 1;
}

/* Internal function: Assigns IP address or range to tmp variable */
static int _ipban_parse(const char *iprange)
{
	char *token = NULL, *range = NULL, *base = NULL;
	unsigned int base2 = 0, range2 = 0, counter = 24;
	char *dup;

	dup = strdup(iprange);
	base = strtok(dup, "/");
	if(base)
	{
		range = strtok(NULL, "/");
		if(range)
		{
			if(strtok(NULL, "/"))
			{
				free(dup);
				return -1;
			}
			range2 = _pow(2, (32 - atoi(range)));
		}

		token = strtok(base, ".");
		while(counter >= 0)
		{
			if(!token) break;
			base2 += atoi(token) << counter;
			counter -= 8;
			token = strtok(NULL, ".");
		}
	}
	else
	{
		free(dup);
		return -1;
	}

	free(dup);

	tmp.base = base2;
	tmp.range = range2;

	return 0;
}

/* Internal function: Add tmp to the ban list*/
static void _ipban_add2()
{
	banlist = (ipban**)realloc(banlist, (bansize + 1) * sizeof(ipban*));
	banlist[bansize] = (ipban*)malloc(sizeof(ipban));
	banlist[bansize]->base = tmp.base;
	banlist[bansize]->range = tmp.range;
	banlist[bansize]->nexttry = time(NULL) + IPBAN_TARPIT;
	banlist[bansize]->banlevel = 1;
	bansize++;
}

/* Internal function: Check whether tmp is in the ban list, or even remove it */
static int _ipban_check2(int remove)
{
	int i;

	for(i = 0; i < bansize; i++)
	{
		if((banlist[i]->base <= tmp.base)
		&& (banlist[i]->base + banlist[i]->range >= tmp.base))
		{
			if(remove)
			{
				banlist[i]->base = 0;
				banlist[i]->range = 0;
				return 1;
			}
			if(time(NULL) < banlist[i]->nexttry) banlist[i]->banlevel++;
			else if(time(NULL) > banlist[i]->nexttry + IPBAN_RESET)
			{
				banlist[i]->banlevel = 0;
				banlist[i]->nexttry = time(NULL) + IPBAN_TARPIT;
			}
			else banlist[i]->nexttry = time(NULL) + IPBAN_TARPIT;
			if(banlist[i]->banlevel >= IPBAN_BANNED) return 1;
			return 0;
		}
	}

	_ipban_add2();

	return 0;
}

/* Public function */
int ipban_add(const char *iprange)
{
	if(!banlist) _ipban_init();

	if(_ipban_parse(iprange) != 0) return -1;
	if(_ipban_check2(0) != 0) return 1;

/*printf("ADD: %u +%u\n", tmp.base, tmp.range);*/
	_ipban_add2();

	banlist[bansize - 1]->nexttry = 0;
	banlist[bansize - 1]->banlevel = 0xFFFFFFFF;

	return 0;
}

/* Public function */
int ipban_remove(const char *iprange)
{
	int ret;

	if(!banlist) _ipban_init();

	if(_ipban_parse(iprange) != 0) return -1;
	ret = _ipban_check2(1);

	return !ret;
}

/* Public function */
int ipban_check(const char *ipaddress)
{
	int ret;

	if(!banlist) _ipban_init();

	if(_ipban_parse(ipaddress) != 0) return -1;
	ret = _ipban_check2(0);

	return ret;
}

